library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;

entity counters is
    port (
        clk133: in STD_LOGIC;
        cycle: in STD_LOGIC_VECTOR (3 downto 0);
        RST: in STD_LOGIC_VECTOR (2 downto 0);
        GC_AUTOMON: in STD_LOGIC;
        UMON_ENA: in STD_LOGIC_VECTOR (2 downto 0);
        UMON_SCLR: in STD_LOGIC_VECTOR (2 downto 0);
-- monitoring counter count inputs
        WCOUNT: in STD_LOGIC_VECTOR(2 downto 0);
        trailer: in STD_LOGIC_VECTOR(2 downto 0);
        ADR: in STD_LOGIC_VECTOR(5 downto 0);
        cntr_data: out STD_LOGIC_VECTOR (31 downto 0)
    );
end counters;

architecture my_arch of counters is
COMPONENT lpm_ram_dp
   GENERIC ( LPM_WIDTH: POSITIVE := 32;
             LPM_WIDTHAD: POSITIVE := 7;
             LPM_NUMWORDS: NATURAL := 128;
             LPM_TYPE: STRING := "LPM_RAM_DP";
             LPM_INDATA: STRING := "REGISTERED";
             LPM_OUTDATA: STRING := "REGISTERED";
             LPM_RDADDRESS_CONTROL: STRING := "REGISTERED";
             LPM_WRADDRESS_CONTROL: STRING := "REGISTERED";
             LPM_FILE: STRING := "lrb_counter.mif";
             LPM_HINT: STRING := "UNUSED"

           );
   PORT    ( rdaddress, wraddress: IN STD_LOGIC_VECTOR(LPM_WIDTHAD-1 DOWNTO 0);
             rdclock, wrclock: IN STD_LOGIC := '0';
             rden, rdclken, wrclken: IN STD_LOGIC := '1';
             wren: IN STD_LOGIC; 
             data: IN STD_LOGIC_VECTOR(LPM_WIDTH-1 DOWNTO 0);
             q: OUT STD_LOGIC_VECTOR(LPM_WIDTH-1 DOWNTO 0));
END COMPONENT;
COMPONENT lpm_add_sub
   GENERIC (LPM_WIDTH: POSITIVE := 16;
      LPM_REPRESENTATION: STRING := "UNSIGNED";
      LPM_DIRECTION: STRING := "ADD";
      LPM_PIPELINE: INTEGER := 1;
      LPM_TYPE: STRING := "LPM_ADD_SUB";
      LPM_HINT: STRING := "UNUSED");
   PORT (dataa, datab: IN STD_LOGIC_VECTOR(LPM_WIDTH-1 DOWNTO 0);
      aclr, clock, cin: IN STD_LOGIC := '0';
      add_sub: IN STD_LOGIC := '1';
      clken: IN STD_LOGIC := '1';
      result: OUT STD_LOGIC_VECTOR(LPM_WIDTH-1 DOWNTO 0);
      cout, overflow: OUT STD_LOGIC);
END COMPONENT;
signal WCOUNT_a : std_logic_vector(2 downto 0);
signal WCOUNT_b : std_logic_vector(2 downto 0);
signal WCOUNT_c : std_logic_vector(2 downto 0);
signal NBLOCK_a : std_logic_vector(2 downto 0);
signal NBLOCK_b : std_logic_vector(2 downto 0);
signal NBLOCK_c : std_logic_vector(2 downto 0);
signal R_WCOUNT_a : std_logic_vector(2 downto 0);
signal R_WCOUNT_b : std_logic_vector(2 downto 0);
signal R_WCOUNT_c : std_logic_vector(2 downto 0);
signal R_NBLOCK_a : std_logic_vector(2 downto 0);
signal R_NBLOCK_b : std_logic_vector(2 downto 0);
signal R_NBLOCK_c : std_logic_vector(2 downto 0);
signal ld_cntr : std_logic_vector(5 downto 0);
signal update_flag : std_logic_vector(5 downto 0);
signal reset_flag : std_logic_vector(5 downto 0);
signal R_update : std_logic_vector(5 downto 0);
signal R_reset : std_logic_vector(5 downto 0);
signal counter_wa : std_logic_vector(6 downto 0);
signal counter_ra : std_logic_vector(6 downto 0);
signal counter_we : std_logic;
signal counter_di : std_logic_vector(31 downto 0);
signal counter_do : std_logic_vector(31 downto 0);
signal write_ena : std_logic;
signal rst_wcount : std_logic;
signal rst_nblock : std_logic;
signal update_wcount : std_logic;
signal update_nblock : std_logic;
signal update_dl : std_logic_vector(1 downto 0);
signal rst_cntr : std_logic;
signal rst_cntr_q : std_logic;
signal cntr_h_q : std_logic_vector(15 downto 0);
signal sum_l : std_logic_vector(16 downto 0);
signal sum_l_cout : std_logic;
signal inc : std_logic_vector(15 downto 0);
signal inc_q : std_logic_vector(2 downto 0);
signal R5xh : std_logic_vector(2 downto 0);
signal R6xh : std_logic_vector(2 downto 0);
signal mux_sel : std_logic_vector(4 downto 0);
begin
cntr_data <= counter_do;
-- counters
process(clk133)
begin
  if(clk133'event and clk133 = '1')then
    if(UMON_SCLR(0) = '1' or ld_cntr(0) = '1')then
      WCOUNT_a <= "00" & WCOUNT(0);
    else
      WCOUNT_a <= WCOUNT_a + ("00" & WCOUNT(0));
    end if;
    if(UMON_SCLR(0) = '1' or ld_cntr(3) = '1')then
      NBLOCK_a <= "00" & trailer(0);
    else
      NBLOCK_a <= NBLOCK_a + ("00" & trailer(0));
    end if;
    if(UMON_SCLR(1) = '1' or ld_cntr(1) = '1')then
      WCOUNT_b <= "00" & WCOUNT(1);
    else
      WCOUNT_b <= WCOUNT_b + ("00" & WCOUNT(1));
    end if;
    if(UMON_SCLR(1) = '1' or ld_cntr(4) = '1')then
      NBLOCK_b <= "00" & trailer(1);
    else
      NBLOCK_b <= NBLOCK_b + ("00" & trailer(1));
    end if;
    if(UMON_SCLR(2) = '1' or ld_cntr(2) = '1')then
      WCOUNT_c <= "00" & WCOUNT(2);
    else
      WCOUNT_c <= WCOUNT_c + ("00" & WCOUNT(2));
    end if;
    if(UMON_SCLR(2) = '1' or ld_cntr(5) = '1')then
      NBLOCK_c <= "00" & trailer(2);
    else
      NBLOCK_c <= NBLOCK_c + ("00" & trailer(2));
    end if;
  end if;
end process;
process(clk133)
begin
  if(clk133'event and clk133 = '1')then
    if(RST(0) = '1')then
      R_WCOUNT_a <= (others => '0');
    elsif(UMON_ENA(0) = '1' or (ld_cntr(0) = '1' and update_flag(0) = '0'))then
      R_WCOUNT_a <= WCOUNT_a;
    end if;
    if(RST(0) = '1')then
      R_NBLOCK_a <= (others => '0');
    elsif(UMON_ENA(0) = '1' or (ld_cntr(3) = '1' and update_flag(3) = '0'))then
      R_NBLOCK_a <= NBLOCK_a;
    end if;
  end if;
end process;
process(clk133)
begin
  if(clk133'event and clk133 = '1')then
    if(RST(1) = '1')then
      R_WCOUNT_b <= (others => '0');
    elsif(UMON_ENA(1) = '1' or (ld_cntr(1) = '1' and update_flag(1) = '0'))then
      R_WCOUNT_b <= WCOUNT_b;
    end if;
    if(RST(1) = '1')then
      R_NBLOCK_b <= (others => '0');
    elsif(UMON_ENA(1) = '1' or (ld_cntr(4) = '1' and update_flag(4) = '0'))then
      R_NBLOCK_b <= NBLOCK_b;
    end if;
  end if;
end process;
process(clk133)
begin
  if(clk133'event and clk133 = '1')then
    if(RST(2) = '1')then
      R_WCOUNT_c <= (others => '0');
    elsif(UMON_ENA(2) = '1' or (ld_cntr(2) = '1' and update_flag(2) = '0'))then
      R_WCOUNT_c <= WCOUNT_c;
    end if;
    if(RST(2) = '1')then
      R_NBLOCK_c <= (others => '0');
    elsif(UMON_ENA(2) = '1' or (ld_cntr(5) = '1' and update_flag(5) = '0'))then
      R_NBLOCK_c <= NBLOCK_c;
    end if;
  end if;
end process;
i_counters: lpm_ram_dp
   port map(
           rdaddress => counter_ra,
           wraddress => counter_wa,
           wrclock => clk133,
           rdclock => clk133,
           wren => counter_we,
           data => counter_di,
           q => counter_do);

counter_wa(5 downto 4) <= "01";
mux_sel(1) <= ld_cntr(2);
mux_sel(3) <= ld_cntr(5);
process(clk133)
begin
  if(clk133'event and clk133 = '1')then
    write_ena <= write_ena or cycle(0);
    counter_we <= not cycle(0) and write_ena;
	counter_wa(6) <= not GC_AUTOMON and not update_dl(1);
    if(cycle(0) = '1')then
      counter_wa(3 downto 0) <= "01" & counter_wa(0) & counter_wa(0);
      counter_ra <= '0' & ADR;
    else
      counter_ra(6) <= not GC_AUTOMON;
      counter_wa(3 downto 0) <= counter_wa(3 downto 0) + 1;
      counter_ra(5 downto 4) <= "01";
      counter_ra(3) <= counter_wa(0) xor not cycle(2);
      counter_ra(2) <= counter_wa(0) xor cycle(2);
      counter_ra(1) <= counter_wa(3) or cycle(3);
      counter_ra(0) <= counter_wa(0);
    end if;
    rst_cntr_q <= rst_cntr;
    inc_q <= inc(2 downto 0);
-- chip can not add 32 bits at 133MHz, must add in two steps
-- synchronous clear of dpram must be done with some tricks to avoid loss of speed
    if(rst_cntr = '1')then
      cntr_h_q <= (others => '0');
    else
      cntr_h_q <= counter_do(31 downto 16);
    end if;
    counter_di(31 downto 16) <= cntr_h_q + ("000000000000000" & (sum_l(16) and not rst_cntr_q));
    if(rst_cntr_q = '1')then
      counter_di(15 downto 0) <= "0000000000000" & inc_q;
    else
      counter_di(15 downto 0) <= sum_l(15 downto 0);
    end if;
	mux_sel(0) <= counter_wa(2);
	mux_sel(2) <= mux_sel(1);
	mux_sel(4) <= mux_sel(3);
    case mux_sel(4 downto 3) is
      when "00" => 
		  R6xh <= R_NBLOCK_a;
		  rst_nblock <= R_reset(3);
		  update_nblock <= R_update(3);
      when "01" =>
		  R6xh <= R_NBLOCK_b;
		  rst_nblock <= R_reset(4);
		  update_nblock <= R_update(4);
      when others =>
		  R6xh <= R_NBLOCK_c;
		  rst_nblock <= R_reset(5);
		  update_nblock <= R_update(5);
    end case;
    case mux_sel(2 downto 1) is
      when "00" =>
		  R5xh <= R_WCOUNT_a;
		  rst_wcount <= R_reset(0);
		  update_wcount <= R_update(0);
      when "01" =>
		  R5xh <= R_WCOUNT_b;
		  rst_wcount <= R_reset(1);
		  update_wcount <= R_update(1);
      when others =>
		  R5xh <= R_WCOUNT_c;
		  rst_wcount <= R_reset(2);
		  update_wcount <= R_update(2);
    end case;
    if(mux_sel(0) = '1')then
      inc(2 downto 0) <= R6xh;
	  rst_cntr <= rst_nblock;
	  update_dl(0) <= update_nblock;
    else
      inc(2 downto 0) <= R5xh;
	  rst_cntr <= rst_wcount;
	  update_dl(0) <= update_wcount;
    end if;
	update_dl(1) <= update_dl(0);
-- load Rx54 thru Rx68 to mux inputs
	ld_cntr(0) <= cycle(0) and counter_wa(0);
	ld_cntr(1) <= ld_cntr(0);
	ld_cntr(2) <= ld_cntr(1);
	ld_cntr(3) <= cycle(0) and not counter_wa(0);
	ld_cntr(4) <= ld_cntr(3);
	ld_cntr(5) <= ld_cntr(4);
	for i in ld_cntr'range loop
-- update flags
	  update_flag(i) <= UMON_SCLR(i mod 3) or (update_flag(i) and not ld_cntr(i));
	  reset_flag(i) <= RST(i mod 3) or (update_flag(i) and ld_cntr(i)) or (reset_flag(i) and not ld_cntr(i));
	  if(ld_cntr(i) = '1')then
		R_reset(i) <= reset_flag(i);
		R_update(i) <= update_flag(i);
	  end if;
	end loop;
  end if;
end process;
inc(15 downto 3) <= "0000000000000";
i_sum_l: lpm_add_sub
   port map(dataa => counter_do(15 downto 0),
            datab => inc,
            clock => clk133,
            result => sum_l(15 downto 0),
            cout => sum_l(16));
end my_arch;
